UNIT EnumeratedTypes_Basic;

INTERFACE

USES
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  System.Generics.Collections, System.TypInfo,
  System.Generics.Defaults, Rtti;

CONST
  UnitID        = 1206;
  Version       = '1.00.00';
  VersionDate   = '2018-03-20';
  DevPlatform   = 'XE7';
  DbPlatform    = 'Not Applicable';
  Author        = 'P.K.Nauta';
  GUID          = '{95CF5AD5-93AD-4261-9611-22142F189CDC}';

TYPE
  TNameCasing       = ( ncNone, ncUpper, ncLower, ncFirst, ncPrefix );
  TNumber           = ( nrOne, nrTwo, nrThree, nrFour );
  TMonth            = ( mtJanuary, mtFebruary, mtMarch, mtApril, mtMay, mtJune );
  TCrState          = ( csRegistered, csAnalysis, csApproved, csDesign, csDevelopment, csTesting, csReleased, csCancelled, csRejected );

  TEnumBasic = CLASS
  PRIVATE
    FDescription    : String;
    FEnumOrd        : Integer;
    FEnumName       : String;
    FName           : String;
  PROTECTED
    FUNCTION        GetCompareValue : String; VIRTUAL; ABSTRACT;
    PROCEDURE       SetEnumOrd( CONST Value : Integer ); VIRTUAL; ABSTRACT;
    PROCEDURE       SetName( CONST Value : String ); VIRTUAL; ABSTRACT;
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    PROPERTY        Description   : String    READ  FDescription    WRITE  FDescription;
    PROPERTY        EnumName      : String    READ  FEnumName;
    PROPERTY        EnumOrd       : Integer   READ  FEnumOrd;
    PROPERTY        Name          : String    READ  FName;
  END;

  TEnumItem< ET > = CLASS( TEnumBasic )       { Create the CLASS structure for an Enumerated Type (ET) }
  PRIVATE
    FOwner          : Pointer;                { We cannot refer to TEnumListExt because it requires TEnumItem to be completely defined }
  PROTECTED
    FEnumerator     : ET;                     { Protected because it might be used by derived types }
    FUNCTION        GetCompareValue : String; OVERRIDE;
    PROCEDURE       SetEnumOrd( CONST Value : Integer ); OVERRIDE;
    PROCEDURE       SetName( CONST Value : String ); OVERRIDE;
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    PROPERTY        Enumerator    : ET        READ  FEnumerator;
    PROPERTY        Owner         : Pointer   READ  FOwner;
  END;

  TCompareValues< EB : TEnumBasic > = CLASS( TComparer< EB > )
  PRIVATE
    FCaseSensitive  : Boolean;
    FSortAscending  : Boolean;
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    FUNCTION        Compare( CONST Left, Right : EB  ) : Integer; OVERRIDE;
    PROPERTY        CaseSensitive  : Boolean  READ  FCaseSensitive    WRITE  FCaseSensitive;
    PROPERTY        SortAscending  : Boolean  READ  FSortAscending    WRITE  FSortAscending;
  END;

  TCompareValuesExt< ET; EE : TEnumItem< ET > > = CLASS( TCompareValues< EE > );

  TEnumBasicList< EB : TEnumBasic, CONSTRUCTOR > = CLASS( TObjectList< EB > )   { Create the CLASS structure for the list of TEnumItems (EI) }
  PRIVATE
    FCaseSensitive  : Boolean;
    FComparer       : TCompareValues< EB >;
    FEnumInfo       : PTypeInfo;
    FNameCasing     : TNameCasing;
    FSortAscending  : Boolean;
  PROTECTED
    FCompareField   : String;
    PROCEDURE       SetCompareField( CONST Field : String ); VIRTUAL; ABSTRACT;
    PROCEDURE       SetNameCasing( CONST Value : TNameCasing ); VIRTUAL;
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    DESTRUCTOR      Destroy; OVERRIDE;
    PROCEDURE       Sort; OVERLOAD;
    PROPERTY        CaseSensitive     : Boolean           READ  FCaseSensitive    WRITE  FCaseSensitive;
    PROPERTY        CompareField      : String            READ  FCompareField     WRITE  SetCompareField;
    PROPERTY        NameCasing        : TNameCasing       READ  FNameCasing       WRITE  SetNameCasing;
    PROPERTY        SortAscending     : Boolean           READ  FSortAscending      WRITE  FSortAscending;
  END;

  TEnumBasicListExt      = CLASS( TEnumBasicList< TEnumBasic > );    { NickName to use as Type Specifier }

  TEnumList< ET; EI : TEnumItem< ET >, CONSTRUCTOR  >  = CLASS( TEnumBasicListExt )         { EE stands for EnumItemExt }
  PRIVATE
    FUNCTION        ConvertEnumToOrd( Value: ET ): Integer;
    FUNCTION        GetEnumItem( Value : ET ) : EI;
    FUNCTION        GetItem( Index : Integer): EI;
    PROCEDURE       Initialize;
  PROTECTED
    PROCEDURE       SetCompareField( CONST Field : String ); OVERRIDE;
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    DESTRUCTOR      Destroy; OVERRIDE;
    FUNCTION        Add( CONST Value : ET ): Integer; REINTRODUCE;
    PROCEDURE       Delete( Index : Integer ); REINTRODUCE;
    FUNCTION        FindEnumName( EnumName : String ) : EI;
    FUNCTION        FindName( Name : String ) : EI;
    FUNCTION        HasProperty( CONST Prop : String ) : Boolean;
    PROCEDURE       ModifyNameDescr( Index : ET; Name : String; Description : String );
    PROCEDURE       PopulateStringList( List : TStrings; Field : String );
    PROPERTY        EnumItems[ Index : ET ]     : EI      READ  GetEnumItem; DEFAULT;    { Setter is not desired because the list has fixed Enumerator elements }
    PROPERTY        Items[ Index : Integer ]    : EI      READ  GetItem;
  END;

  TNumberItem       = CLASS( TEnumItem< TNumber > )
  PRIVATE
  PUBLIC
    PROPERTY        Number      : TNumber   READ  FEnumerator;
  END;

  TNumberItems      = CLASS( TEnumList< TNumber, TNumberItem > )
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
  END;

  TMonthItem        = CLASS( TEnumItem< TMonth > )
  PRIVATE
    FQuarter        : Integer;
  PROTECTED
    PROCEDURE       SetEnumOrd( CONST Value : Integer ); OVERRIDE;
  PUBLIC
    PROPERTY        Quarter     : Integer   READ  FQuarter;
  END;

  TMonthItems       = CLASS( TEnumList< TMonth, TMonthItem > )
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
  END;

  TCrStateItem      = CLASS( TEnumItem< TCrState > )
  PRIVATE
    FOpenState      : Boolean;
  PUBLIC
    PROPERTY        CrState     : TCrState    READ  FEnumerator   WRITE  FEnumerator;
    PROPERTY        OpenState   : Boolean     READ  FOpenState    WRITE  FOpenState;
  END;

  TCrStateList      = CLASS( TEnumList< TCrState, TCrStateItem > )
  PRIVATE
  PUBLIC
    CONSTRUCTOR     Create; OVERLOAD;
    PROCEDURE       PopulateOpenStates( List : TStrings );
  END;

FUNCTION ConvertNameCasing( Name : String; Casing : TNameCasing; PrefixSize : Integer = 0 ) : String;
FUNCTION ConvertValueToString( AProperty : TRttiProperty; Value : TValue; ForSorting : Boolean ) : String;

{===========================================================================}
IMPLEMENTATION

USES Vcl.Dialogs;
{===========================================================================}

{ TEnumItem }

{===========================================================================}
CONSTRUCTOR TEnumBasic.Create;

BEGIN
  INHERITED;
END;
{===========================================================================}

{ TEnumItem<T> }

{===========================================================================}
CONSTRUCTOR TEnumItem< ET >.Create;

BEGIN
  INHERITED;
END;
{===========================================================================}
FUNCTION TEnumItem< ET >.GetCompareValue : String;
{                                                                           }
{   This function gets the value for the CompareField. As the RTTI sees     }
{   this as an TEnumItem<ET> and not as an item derived from TEnumItem<ET>, }
{   all additional properties in derived items will lead to errors like     }
{   'Property XXXX does not exist', when calling                            }
{      sField        := TEnumBASICListExt( FOwner ).FCompareField;          }
{      Result        := GetPropValue( Self, sField );                       }
{   Therefore we resolve this via RTTI, which has knowledge on the actual   }
{   derived TEnumItem.                                                      }
{                                                                           }
VAR
  rContext  : TRttiContext;
  rType     : TRttiType;
  rProperty : TRttiProperty;
  oValue    : TValue;

BEGIN
  rType     := rContext.GetType( ClassType );
  rProperty := rType.GetProperty( TEnumBasicListExt( FOwner ).FCompareField );
  oValue    := rProperty.GetValue( Self );
  Result    := ConvertValueToString( rProperty, oValue, True );
END;
{===========================================================================}
PROCEDURE TEnumItem< ET >.SetEnumOrd( CONST Value : Integer );

BEGIN
  FEnumName   := GetEnumName( TypeInfo( ET ), Value );
  FEnumOrd    := Value;
  FEnumerator := TRttiEnumerationType.GetValue<ET>( FEnumName );
                  { FEnumerator := ET( FEnumOrd ) does not compile:  E2089 Invalid Typecast }
END;
{===========================================================================}
PROCEDURE TEnumItem< ET >.SetName( CONST Value : String );

BEGIN
  FName  := ConvertNameCasing( Value, TEnumBasicListExt( FOwner ).NameCasing );
END;
{===========================================================================}

{ TEnumList<TEnumItem> }

{===========================================================================}
CONSTRUCTOR TEnumBasicList< EB >.Create;

BEGIN
  INHERITED Create;

  FCompareField   := 'NAME';
  FCaseSensitive  := False;
  FComparer       := TCompareValues< EB >.Create;
  FEnumInfo       := NIL;     { To be set by derived items }
  FNameCasing     := ncNone;
  FSortAscending    := True;
END;
{===========================================================================}
DESTRUCTOR TEnumBasicList< EB >.Destroy;

BEGIN
  FComparer.Free;

  INHERITED Destroy;
END;
{===========================================================================}
PROCEDURE TEnumBasicList< EB >.SetNameCasing( CONST Value : TNameCasing );

VAR
  iEnum         : Integer;

BEGIN
  FNameCasing   := Value;

  FOR iEnum := 0 TO Count - 1 DO
    Items[ iEnum ].SetName( Items[ iEnum ].Name );     { This will call SetName, which sets the NameCasing }
END;
{===========================================================================}
PROCEDURE TEnumBasicList< EB >.Sort;

BEGIN
  FComparer.FCaseSensitive    := FCaseSensitive;
  FComparer.FSortAscending      := FSortAscending;

  INHERITED Sort( FComparer AS IComparer< EB > );
END;
{===========================================================================}

{ TEnumListExt< ET, EI > }

{===========================================================================}
CONSTRUCTOR TEnumList< ET, EI >.Create;

BEGIN
  INHERITED Create;

  FEnumInfo       := TypeInfo( ET );
  Initialize;
END;
{===========================================================================}
DESTRUCTOR TEnumList< ET, EI >.Destroy;

BEGIN
  INHERITED Destroy;
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.Add( CONST Value : ET ) : Integer;
{                                                                           }
{   This procedure does nothing as it is not allowed to delete or add items }
{   to the list.                                                            }
{                                                                           }
BEGIN
  RAISE Exception.Create( 'Addition is not allowed' );
END;
{===========================================================================}
PROCEDURE TEnumList< ET, EI >.Delete( Index : Integer );
{                                                                           }
{   This procedure does nothing as it is not allowed to delete or add items }
{   to the list.                                                            }
{                                                                           }
BEGIN
  RAISE Exception.Create( 'Delete is not allowed' );
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.ConvertEnumToOrd( Value: ET ): Integer;
{
{   This function converts an Enumerated Value into its corresponding Order.}
{   One would expect to use                                                 }
{       Result   := Ord( Value );                                           }
{   But this results in compiler error E2008: Incompatible Types. The cause }
{   for this is the situation that the compiler does not know that ET is an }
{   Enumerated Type. The solution was found in the Rtti Unit.               }
{                                                                           }
BEGIN
  CASE System.TypInfo.GetTypeData( TypeInfo( ET ) )^.OrdType OF
    otUByte, otSByte :  Result := PByte( @Value )^;
    otUWord, otSWord :  Result := PWord( @Value )^;
    otULong, otSLong :  Result := PInteger( @Value )^;
  END;
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.FindEnumName( EnumName : String ) : EI;

VAR
  iEnum   : Integer;

BEGIN
  Result  := NIL;

  FOR iEnum := 0 TO Count - 1 DO
    IF ( Items[ iEnum ].EnumName = EnumName ) THEN
      BEGIN
        Result    := Items[ iEnum ];
        Break;
      END;
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.FindName( Name : String ) : EI;

VAR
  iEnum     : Integer;
  iCompare  : Integer;

BEGIN
  Result    := NIL;

  FOR iEnum := 0 TO Count - 1 DO
    BEGIN
      iCompare    := AnsiCompareText( Items[ iEnum ].Name, Name );

      IF ( iCompare = 0 ) THEN
        BEGIN
          Result  := Items[ iEnum ];
          Break;
        END;
    END;
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.GetEnumItem( Value : ET ) : EI;

VAR
  iOrd    : Integer;
  iEnum   : Integer;

BEGIN
  iOrd    := ConvertEnumToOrd( Value );

  FOR iEnum := 0 TO Count - 1 DO
    BEGIN
      IF ( Items[ iEnum ].EnumOrd = iOrd )THEN
        BEGIN
          Result    := Items[ iEnum ];
          Break;
        END;
    END;
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.GetItem( Index : Integer ) : EI;

BEGIN
  Result   := EI( INHERITED Items[ Index ] );
END;
{===========================================================================}
FUNCTION TEnumList< ET, EI >.HasProperty( CONST Prop : String ) : Boolean;

VAR
  rContext  : TRttiContext;
  rType     : TRttiType;
  rProperty : TRttiProperty;

BEGIN
  rType     := rContext.GetType( TypeInfo( EI ) );
  rProperty := rType.GetProperty( Prop );
  Result    := ( rProperty <> NIL );
END;
{===========================================================================}
PROCEDURE TEnumList< ET, EI >.Initialize;
{                                                                           }
{   This procedure adds a new Enumerated Item to the list, for exactly as   }
{   many enumerators that are present in the Enumerated Type. It will only  }
{   set the Creator, Enumerator, its associated EnumOrd and EnumName. It    }
{   will also set the Name equal to the EnumName, as default but unique     }
{   value. Further detail settings need to be arranged outside this CLASS.  }
{                                                                            }
VAR
  iEnumOrd    : Integer;
  rEnumItem   : EI;
  rEnumType   : ET;

BEGIN
  FOR iEnumOrd := FEnumInfo.TypeData.MinValue TO FEnumInfo.TypeData.MaxValue DO
    BEGIN
      rEnumItem              := EI.Create;
      rEnumItem.FOwner       := Self;
      rEnumItem.SetEnumOrd( iEnumOrd );         { This will also set FEnumName and FEnumerator }
      rEnumItem.SetName( rEnumItem.FEnumName ); { Make Name as default equal to EnumName       }

      INHERITED Add( rEnumItem );
  END;
END;
{===========================================================================}
PROCEDURE TEnumList< ET, EI >.ModifyNameDescr( Index : ET; Name : String; Description : String );

VAR
  rItem       : EI;
  rExisting   : EI;

BEGIN
  rExisting   := FindName( Name );  { Check if Name is already in use }

  IF ( rExisting = NIL ) THEN
    BEGIN
      rItem               := GetEnumItem( Index );
      rItem.SetName( Name );
      rItem.Description   := Description;
    END
  ELSE
    RAISE Exception.Create( 'Name ''' + Name + ''' is already in use with item ' + rExisting.EnumName );
END;
{===========================================================================}
PROCEDURE TEnumList< ET, EI >.PopulateStringList( List : TStrings; Field : String );

VAR
  rContext    : TRttiContext;
  rType       : TRttiType;
  rProperty   : TRttiProperty;
  oValue      : TValue;
  iEnum       : Integer;

BEGIN
  List.Clear;
  rType       := rContext.GetType( TypeInfo( EI ) );
  rProperty   := rType.GetProperty( Field );

  FOR iEnum := 0 TO Count - 1 DO
    BEGIN
      oValue  := rProperty.GetValue( TObject( Items[ iEnum ] ) );    { Casting to TObject is needed to avoid compiler problems like E2010: Incompatible type: 'Pointer' and EI }
      List.Add( ConvertValueToString( rProperty, oValue, False ) );
    END;
END;
{===========================================================================}
PROCEDURE TEnumList< ET, EI >.SetCompareField( CONST Field : String );

BEGIN
  IF HasProperty( Field ) AND ( UpperCase( Field ) <> 'COMPAREFIELD' ) THEN
    FCompareField   := UpperCase( Field )             { Comparefield is a calculated field, so we cannot compare on this field }
  ELSE
    RAISE Exception.Create( 'Field does not exist: ' + Field );
END;
{===========================================================================}

{ TCompareValues< EB > }

{===========================================================================}
CONSTRUCTOR TCompareValues< EB >.Create;

BEGIN
  INHERITED Create;

  FCaseSensitive  := False;
  FSortAscending    := True;
END;
{===========================================================================}
FUNCTION TCompareValues< EB >.Compare( CONST Left, Right : EB ) : Integer;
{                                                                           }
{   This function compares two CompareValues for sorting purposes.          }
{   The CompareValue is a calculated field, representing FCompareField.     }
{                                                                           }
BEGIN
  IF FCaseSensitive THEN
    Result    := AnsiCompareStr( Left.GetCompareValue, Right.GetCompareValue )
  ELSE
    Result    := AnsiCompareText( Left.GetCompareValue, Right.GetCompareValue );

  IF NOT FSortAscending THEN
    Result    := - Result;
END;
{===========================================================================}

{ TMonthItem }

{===========================================================================}
PROCEDURE TMonthItem.SetEnumOrd( CONST Value : Integer );

BEGIN
  INHERITED SetEnumOrd( Value );

  FQuarter    := 1 + ( EnumOrd DIV 3 );
END;
{===========================================================================}

{ TMonthItems }

{===========================================================================}
CONSTRUCTOR TMonthItems.Create;

BEGIN
  INHERITED;

  ModifyNameDescr( mtJanuary,  'JAN',    'January' );
  ModifyNameDescr( mtFebruary, 'FEB',    'February' );
  ModifyNameDescr( mtMarch,    'MAR',    'March' );
  ModifyNameDescr( mtApril,    'APR',    'April' );
  ModifyNameDescr( mtMay,      'MAY',    'May' );
  ModifyNameDescr( mtJune,     'JUN',    'June' );
END;
{===========================================================================}

{ TNumberItems }

{===========================================================================}
CONSTRUCTOR TNumberItems.Create;

BEGIN
  INHERITED;

  FNameCasing   := ncUpper;

  ModifyNameDescr( nrOne,   'ONE',    'OneOneOne' );
  ModifyNameDescr( nrTwo,   'TWO',    'TwoTwoTwo' );
  ModifyNameDescr( nrThree, 'THREE',  'ThreeThreeThree' );
  ModifyNameDescr( nrFour,  'four',   'FourFourFour' );
END;
{===========================================================================}

{ TCrStateList }

{===========================================================================}
CONSTRUCTOR TCrStateList.Create;

BEGIN
  INHERITED Create;

  ModifyNameDescr( csRegistered,  'Registered',  'The CR was registered but needs Analysis' );
  ModifyNameDescr( csAnalysis,	   'Analysis',    'The CR is being analyzed' );
  ModifyNameDescr( csApproved,	   'Approved',    'The CR is approved, budget available, work can start' );
  ModifyNameDescr( csDesign,      'Design',      'The Design for the CR is going on' );
  ModifyNameDescr( csDevelopment, 'Development', 'The Code Development was started' );
  ModifyNameDescr( csTesting,	   'Testing',     'The Code is being Tested' );
  ModifyNameDescr( csReleased, 	 'Released',    'The Code was Released into Production' );
  ModifyNameDescr( csCancelled, 	 'Cancelled',   'Work on the CR was stopped' );
  ModifyNameDescr( csRejected,	   'Rejected',    'After Analysis, the CR was rejected' );

  EnumItems[ csRegistered ].OpenState   := True;
  EnumItems[ csAnalysis ].OpenState     := True;
  EnumItems[ csApproved ].OpenState     := True;
  EnumItems[ csDesign ].OpenState       := True;
  EnumItems[ csDevelopment ].OpenState  := True;
  EnumItems[ csTesting ].OpenState      := True;
END;
{===========================================================================}
PROCEDURE TCrStateList.PopulateOpenStates( List : TStrings );

VAR
  iEnum       : Integer;

BEGIN
  List.Clear;

  FOR iEnum := 0 TO Count - 1 DO
    IF Items[ iEnum ].OpenState THEN
      List.Add( Items[ iEnum ].Name );
END;
{===========================================================================}

{ Generic Functions }

{===========================================================================}
FUNCTION ConvertNameCasing( Name : String; Casing : TNameCasing; PrefixSize : Integer = 0 ) : String;

BEGIN
  CASE Casing OF
    ncUpper     : Result  := UpperCase( Name );
    ncLower     : Result  := LowerCase( Name );
    ncFirst     : Result  := UpperCase( Copy( Name, 1, 1 ) ) + LowerCase( Copy( Name, 2, Length( Name ) ) );
    ncPrefix    : Result  := LowerCase( Copy( Name, 1, PrefixSize ) ) + UpperCase( Copy( Name, PrefixSize + 1, 1 ) ) + LowerCase( Copy( Name, PrefixSize + 2, Length( Name ) ) );
  ELSE
                  Result  := Name;
  END;
END;
{===========================================================================}
FUNCTION ConvertValueToString( AProperty : TRttiProperty; Value : TValue; ForSorting : Boolean ) : String;
{===========================================================================}
  FUNCTION IsBoolType( ATypeInfo : PTypeInfo ) : Boolean;
  BEGIN
    Result := ( ATypeInfo = System.TypeInfo( Boolean ) ) OR
              ( ATypeInfo = System.TypeInfo( ByteBool ) ) OR
              ( ATypeInfo = System.TypeInfo( WordBool ) ) OR
              ( ATypeInfo = System.TypeInfo( LongBool ) );
  END;
{===========================================================================}
VAR
  iAdd        : Integer;
  rPropInfo   : PTypeInfo;

BEGIN
  CASE AProperty.PropertyType.TypeKind OF
    tkEnumeration :
      BEGIN
        rPropInfo   := AProperty.PropertyType.Handle;

        IF IsBoolType( rPropInfo ) THEN { Booleans have TypeKind = tkEnumerated }
          IF ( Value.AsOrdinal = 0 )  THEN  Result := 'False'  ELSE  Result := 'True'
        ELSE
          Result    := GetEnumName( rPropInfo, Value.AsOrdinal );  { Return the EnumName }
      END;
    tkInteger :
      BEGIN
        iAdd        := 0;
        IF ForSorting THEN iAdd := 1000000;
        Result      := IntToStr( Value.AsInteger + iAdd );        { To avoid compare problems like 2 > 11 }
      END;
  ELSE
        Result      := Value.AsString;
  END;
END;
{===========================================================================}
END.

